/*****************************************************************************
 * File:            fm10k-dump.c
 * Creation Date:   December 16, 2020
 * Description:     Utility to dump Non Volatile Memory on SPI for FM10K devices. Requires FM10K kernel module running with UIO enabled.
 * Compile:         gcc -o fm10k-dump fm10k-dump.c
 * Usage:           ./fm10k-dump </dev/uio0> <output.bin> [readSizeInMegabytes=1]
 *
 * Copyright (c) 2014 - 2015, Intel Corporation
 * Copyright (c) 2020, FM10K-Documentation Contributors
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright notice,
 *       this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Intel Corporation nor the names of its contributors
 *       may be used to endorse or promote products derived from this software
 *       without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *****************************************************************************/

#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>

#define FM10K_UIO_SIZE 0x0000000004000000
#define FM10K_UIO_OFFSET 0

//From Datasheet, 11.25.1 MGMT Map, Table 11-37
#define FM10K_REGISTER_BASE 0

#define FM10K_REGISTER_SPI_TX_DATA ((0xC26) + (FM10K_REGISTER_BASE))
#define FM10K_REGISTER_SPI_RX_DATA ((0xC27) + (FM10K_REGISTER_BASE))
#define FM10K_REGISTER_SPI_HEADER ((0xC28) + (FM10K_REGISTER_BASE))
#define FM10K_REGISTER_SPI_CTRL   ((0xC29) + (FM10K_REGISTER_BASE))

#define FM10K_SPI_FREQ_KHZ 50000



void ReadRegister32(uintptr_t mem, uint32_t addr, uint32_t* value) {
    *value = *(((uint32_t*)mem) + addr);
}

void WriteRegister32(uintptr_t mem, uint32_t addr, uint32_t value) {
    *(uint32_t volatile*)(((uint32_t*)mem) + addr) = value;

    //If returning instantly this will caused missed writes. nanosleep wait of 1 nsec was too much. Busy wait works
    volatile uint32_t dummy;
    for(uint32_t i = 0; i < 100; ++i){
        dummy = i;
    }
}

uint32_t get_interval_diff(struct timeval *begin, struct timeval *end) {
    struct timeval endT;
    struct timeval diff;

    if (end == NULL) {
        gettimeofday(&endT, NULL);
    } else {
        endT.tv_sec = end->tv_sec;
        endT.tv_usec = end->tv_usec;
    }

    diff.tv_sec  = endT.tv_sec - begin->tv_sec;
    diff.tv_usec = endT.tv_usec - begin->tv_usec;

    return (diff.tv_sec * 1000 + diff.tv_usec / 1000);
}

uint32_t fm10k_uio_spi_SetCtrlReg(uintptr_t mem, uint32_t value){
    uint32_t spiCtrl;
    struct timeval startTime;
    uint8_t isTimeout = 0;

    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, value);

    gettimeofday(&startTime, NULL);
    do{
        if(isTimeout){
            printf("Timeout waiting for SPI_CTRL Busy. 0x%02x\n", spiCtrl);
            return 1;
        }

        if(get_interval_diff(&startTime, NULL) > 50){
            isTimeout = 1;
        }

        ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl);

    } while ((spiCtrl >> 21) & 1); //While "Busy"

    /* write back SPI_CTRL with command = 0 */
    spiCtrl &= 0xffff87ff;
    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, spiCtrl);

    return 0;
}

void fm10k_uio_spi_Enable(uintptr_t mem){
    //ENABLE SPI
    uint32_t spiCtrl = 0;
    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl);
    /* keep current freq setting and set SPI Enable */
    spiCtrl &= 0x3ff;
    spiCtrl |= 1<<10;
    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, spiCtrl);
}

void fm10k_uio_spi_Disable(uintptr_t mem){
    //DISABLE SPI
    uint32_t spiCtrl = 0;
    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl);
    /* keep current freq setting and set SPI Disable */
    spiCtrl &= 0x3ff;
    WriteRegister32(mem, FM10K_REGISTER_SPI_CTRL, spiCtrl);
}

uint32_t fm10k_uio_spi_ReadFlash(uintptr_t mem, uint32_t address, uint8_t* data, int32_t len){

    uint32_t spiCtrl;
    uint32_t rxData;
    uint32_t header;
    int32_t freq;
    int32_t cnt;
    int32_t numRead;

    fm10k_uio_spi_Enable(mem);


    ReadRegister32(mem, FM10K_REGISTER_SPI_CTRL, &spiCtrl);

    if(FM10K_SPI_FREQ_KHZ > 0){
        freq = ((100000 / (int)FM10K_SPI_FREQ_KHZ) / 2) - 1;
        if (freq < 0){
            freq = 0;
        }

        spiCtrl ^= ((spiCtrl >> 0) ^ freq) & ((2 << 9) - 1); //Set Freq
    }

    spiCtrl &= 0x7ff;

    /* header: command (1 byte: READ_BYTES) + address (3 bytes) */
    header = (0x3 << 24) | (address & 0xffffff);
    WriteRegister32(mem, FM10K_REGISTER_SPI_HEADER, header);

    /* first loop only: set send header flag. Following loops: shift only data. */
    spiCtrl |= (0x1 << 11);

    cnt = 0;
    while (cnt < len){
        /* determine the number of data bytes to read [1..4] */
        numRead = (len - cnt) > 3 ? 4 : (len - cnt);
        /* set 'shift data' flag and number of data bytes */
        spiCtrl |= ((0x4 << 11) | ((numRead & 0x03) << 17));

        if(fm10k_uio_spi_SetCtrlReg(mem, spiCtrl)){
            return 1;
        }

        /* get data */
        ReadRegister32(mem, FM10K_REGISTER_SPI_RX_DATA, &rxData);

        /* push the read data into the array */
        while (numRead) {
            numRead--;
            data[cnt++] = (rxData >> (numRead * 8)) & 0xff;
        }

        spiCtrl &= 0x7ff;
    }

    /* release CS */
    spiCtrl |= (0x8 << 11);
    if(fm10k_uio_spi_SetCtrlReg(mem, spiCtrl)){
        return 1;
    }


    fm10k_uio_spi_Disable(mem);

    return 0;
}

int main(int argc, char** argv){
    int fdUIO;
    FILE* fdOutput;

    if(argc < 3){
        printf("Usage: %s </dev/uio0> <output.bin> [sizeInMB, default 1]\n", argv[0]);
        return 1;
    }

    fdOutput = fopen(argv[2], "wb");

    if(fdOutput == NULL){
        printf("Could not create %s\n", argv[2]);
        return 1;
    }

    fdUIO = open(argv[1], O_RDWR);
    if(fdUIO <= 0){
        printf("Unable to open uio device %s to read NVM\n", argv[1]);
        return 1;
    }
    void* memmapAddr;
    memmapAddr = mmap(NULL, FM10K_UIO_SIZE, PROT_READ | PROT_WRITE, MAP_SHARED, fdUIO, 0);
    if (memmapAddr == MAP_FAILED) {
        printf("Unable to map uio device %s to read NVM\n", argv[1]);
        close(fdUIO);
        return 1;
    }

    unsigned int mbToRead = 1;

    if(argc > 3){
        sscanf(argv[3], "%u", &mbToRead);
    }

    if(mbToRead > 128){
        printf("Too many MB: %u\n", mbToRead);
        return 1;
    }

    uint32_t bootImageSize = mbToRead * 1024 * 1024;

    uint8_t* value = malloc(bootImageSize);

    uint32_t strideSize = 1024;

    for(uint32_t addr = 0; addr < bootImageSize; addr += strideSize){
        printf("\rLOC 0x%08x MAX 0x%08x", addr, bootImageSize);
        if(fm10k_uio_spi_ReadFlash((uintptr_t)(uint32_t *)memmapAddr, addr, value + addr, strideSize)){
            return 2;
        }
    }


    fwrite(value, 1, bootImageSize, fdOutput);

    free(value);

    fclose(fdOutput);

    printf("\nRead %u bytes\n", bootImageSize);

    return 0;
}